package samples;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import javax.ws.rs.core.Response.Status;

import unlimited.fc.rest.shared.model.FileObjectModel;
import unlimited.fc.rest.shared.model.FileObjectsModel;
import unlimited.fc.rest.transferagent.model.ConnectionKeyModel;
import unlimited.fc.rest.transferagent.model.TransferFilesRequestModel;
import unlimited.fc.rest.transferagent.model.TransferJobStatusDetailsModel;
import unlimited.fc.rest.transferagent.model.TransferStatusDetailsModel;
import util.RESTEvoker;

/**
 * Base class instance that is used to provide common functionality across all TransferAgent REST API examples.
 * Resource contains common methods for creating connectionKeys (used when talking to remoteServer), creating 
 * transfer requests, and for outputting transfer information
 */
public abstract class SampleBase implements SampleConstants {

  protected RESTEvoker restEvoker;
  
 /**
  * Generates a connectionKey that will be needed for any request that communicates with the Server
  * @param server IP of the Server
  * @param port Port of the Server
  * @param username Username of the user account 
  * @param password Password of the user account
  * @return connectionKey hash that will be used to perform future operations with the TA
  * @throws Exception
  */
 protected String generateConnectionKey(String server, String port, String username, String password) throws Exception {
   restEvoker.addQueryParam("remoteServer", server);
   restEvoker.addQueryParam("remotePort", String.valueOf(port));
   restEvoker.addQueryParam("username", username);
   restEvoker.addQueryParam("password", password);
   
   final String connectRESTPath = getRESTURLForPath(SampleConstants.CONNECT_PATH);
   final ConnectionKeyModel connectKeyResp = restEvoker.doGETCall(connectRESTPath, ConnectionKeyModel.class, Status.OK);
   return connectKeyResp.getConnectionKey();
 }
 
 /**
  * Helper method that generates basic TransferRequest objects. 
  * 
  * Note: If the sourceConnectionKey is set to "local" then the TA will be executing an upload of the files set. 
  * If the destinationConnectionKey is set to "local" then the TA will be executing a download from the Server
  * 
  * @param sourceConnectionKey The source of the files that will be transferred. 
  * @param destinationConnectionKey The destination of the files that will be transferred. If set to "local", then TA will download remote files that are provided
  * @return A basic transfer model that can be submitted to the TransferAgent
  */
 protected TransferFilesRequestModel getTransferFilesModel(String sourceConnectionKey, String destinationConnectionKey) {
   TransferFilesRequestModel request = new TransferFilesRequestModel();
   request.setSourceConnectionKey(sourceConnectionKey);
   request.setDestinationConnectionKey(destinationConnectionKey);
   return request;
 }

  /**
   * Submits the transfer to the TransferAgent and awaits for it to complete
   * 
   * @param filesModel UploadTransferRequest containing the data needed for the transfer
   * @throws Exception Thrown if the TransferAgent encounters an error during the transaction
   */
  protected void performTAFileTransfer(TransferFilesRequestModel filesModel) throws Exception {
    final String transferRESTPath = getRESTURLForPath(SampleConstants.TRANSFER_PATH);
    final TransferFilesRequestModel transferRequest = restEvoker.doPOSTCall(transferRESTPath, filesModel, TransferFilesRequestModel.class, Status.OK);
    System.out.println("Request accepted. Awaiting transfer completion");

    // Transfer has started, wait for it to finish
    final String transferJobID = transferRequest.getJobId();
    awaitTransferCompletion(transferJobID);

    // Output some statistical information
    final String restURL = getRESTURLForPath(SampleConstants.STATUS_PATH + "/" + transferJobID);
    TransferStatusDetailsModel currentStatus = restEvoker.doGETCall(restURL, TransferStatusDetailsModel.class, Status.OK);
    outputTransferData(currentStatus);
  }

  /**
   * Simple method that will await until the transfer is complete, or cancelled
   * 
   * @param transferID ID of the transfer that you are interested in
   * @throws Exception
   */
  protected void awaitTransferCompletion(String transferID) throws Exception {
    final String restURL = getRESTURLForPath(SampleConstants.STATUS_PATH + "/" + transferID);
    TransferStatusDetailsModel currentStatus = restEvoker.doGETCall(restURL, TransferStatusDetailsModel.class, Status.OK);

    while (!currentStatus.getTransferComplete() && !currentStatus.getTransferCancelled()) {
      Thread.sleep(500);
      outputTransferData(currentStatus);

      currentStatus = restEvoker.doGETCall(restURL, TransferStatusDetailsModel.class, Status.OK);
    }

    outputTransferData(currentStatus);
  }
  
  
  /**
   * Outputs file data in a friendly fashion when called. 
   * @param fileData File listing that you wish to see the contents for
   */
  protected void outputFileListData(FileObjectsModel fileData) {
    final List<FileObjectModel> innerFileList = fileData.getFileObjectList();

    if (innerFileList != null) {
      System.out.println("Here is the listing for dir (" + fileData.getDirectory() + ")");
      for (FileObjectModel file : innerFileList) {
        System.out.format("%-30s%-30s%n", "  ---> File name", file.getName());
        System.out.format("%-30s%-30s%n", "  ---> File path", file.getPath());
        System.out.format("%-30s%-30s%n", "  ---> File size", file.getSize());
        System.out.format("%-30s%-30s%n%n", "  ---> Is directory?", file.getIsDirectory());
      }
    }
  }
 
   /**
    * Simple method that outputs the current TransferData
    * @param currentStatus
    */
   protected void outputTransferData(TransferStatusDetailsModel currentStatus) {
     final List<TransferJobStatusDetailsModel> currentTransfers = currentStatus.getFileTransfers();
     List<String> currentFilenames = (currentTransfers != null) ? currentTransfers.stream()
                                                                                  .map(transfer -> transfer.getCurrentFilename())
                                                                                  .collect(Collectors.toList()) 
                                                                : new ArrayList<>();
     
     System.out.println("\nTransfer Data:");
     System.out.format("%-30s%-30s%n", "  ---> Transfer Direction", currentStatus.getTransferDirection());
     System.out.format("%-30s%-30s%n", "  ---> Current Percentage", currentStatus.getPercent() + "%");
     System.out.format("%-30s%-30s%n", "  ---> Current Transfers", currentFilenames);
     System.out.format("%-30s%-30s%n", "  ---> Transfer Errors", currentStatus.getTransferErrors().toString());
   }
   
   
   /******************************************************************************************
    *                                 ABSTRACT METHODS
    *******************************************************************************************/
   public abstract String getRESTURLForPath(String restPath);
}
