package samples.transferAgent;
import java.net.URI;
import java.util.Arrays;
import java.util.List;

import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.UriBuilder;

import unlimited.fc.rest.shared.model.FileObjectsModel;
import unlimited.fc.rest.transferagent.model.TransferFilesRequestModel;

/**
 * Small example that is used to outline the external filesystem (EFS) functionality present within the TransferAgent.
 * Functionality supports listings of EFS systems, uploads from EFS storage, and downloads to EFS.
 * 
 * Sample prerequisite:
 *    --- Extract the "fc_EFS_drivers.zip" distribution into the FileCatalystTransferAgent's install directory.
 * 
 *    --- Update the Server connection settings in TransferAgentSampleConstants.java to point to the Server you wish to connect to
 *    --- Update the EFS sample constants to point to a currently supported EFS storage
 *    --- Update FileTransferConstants to match 
 */
public class TAEFSSample extends TASampleBase {

  private URI sampleEFSBaseURI;
    
  // EFS Sample Constants
  private final static String EFS_STORAGE_SCHEME = "smb";
  private final static String EFS_STORAGE_USERNAME = "sampleSambaUsername";
  private final static String EFS_STORAGE_PASSWORD = "sampleSambaPassword";
  private final static String EFS_STORAGE_HOST = "127.0.0.1";
  
  // File Transfer Constants
  private final static String SERVER_FILE_TO_DOWNLOAD = "sampleDownloadFile.txt";
  private final static String EFS_EXAMPLE_UPLOAD_FILE = "exampleShare/testFile.txt";
  private final static String EFS_EXAMPLE_DOWNLOAD_LOCATION = "exampleShare/downloadLocation";
  
  public TAEFSSample() {
    super();
    this.sampleEFSBaseURI = generateEFSFileURL("");
  }
  
  /**
   * Generates the file URL object that all external file system paths will extend off of 
   * @param filePath Path of the file on the external file system resource
   * @return File URI representing the path to the excternal file system storage
   */
  private URI generateEFSFileURL(String filePath) {
    String efsUserInfo = EFS_STORAGE_USERNAME + ":" + EFS_STORAGE_PASSWORD;
    final UriBuilder uriBuilder = UriBuilder.fromPath(filePath)
                                            .scheme(EFS_STORAGE_SCHEME)
                                            .userInfo(efsUserInfo)
                                            .host(EFS_STORAGE_HOST);
    return uriBuilder.build();
  } 
  
  /**
   * Short example method that performs a local file listing of the currently configured EFS storage
   * @throws Exception Thrown if the TransferAgent encounters an error, or the response code is invalid 
   */
  public void performEFSListingExample() throws Exception {
    System.out.println("Performing sample EFS listing for " + sampleEFSBaseURI.toString());
    
    String efsFileListingPath = getRESTURLForPath(LOCAL_PATH + "/" + sampleEFSBaseURI);
    final FileObjectsModel responseData = restEvoker.doGETCall(efsFileListingPath, FileObjectsModel.class, Status.OK);
    outputFileListData(responseData);
  }
  
  /**
   * Example method that performs an upload from an external filesystem, and to the connected Server
   * @throws Exception Thrown if a connection is not possible, or if the sample encounters an issue when communicating with the TransferAgent
   */
  public void performEFSUploadExample() throws Exception {
    URI efsFileURIPath = generateEFSFileURL(EFS_EXAMPLE_UPLOAD_FILE);
    System.out.println("\nPerforming EFS upload sample. Uploading file (" + efsFileURIPath +")");
    
    // Generate the file transfer request that we will be uploading to the system
    String connectionKey = generateConnectionKey(SERVER_IP, SERVER_PORT_STR, SERVER_USERNAME, SERVER_PASSWORD);
    final TransferFilesRequestModel efsUploadTransferRequest = getTransferFilesModel("local", connectionKey);
    List<String> filesToUpload = Arrays.asList(efsFileURIPath.toString());
    efsUploadTransferRequest.setFileList(filesToUpload);
    
    // Perform the transfer
    performTAFileTransfer(efsUploadTransferRequest);
    System.out.println("Upload operation has completed.");
  }
  
  /**
   * Example method that downloads a file from the Server, and places it within the external file storage
   * @throws Exception 
   */
  public void performEFSDownloadExample() throws Exception {
    URI efsFileURIPath = generateEFSFileURL(EFS_EXAMPLE_DOWNLOAD_LOCATION);
    System.out.println("\nPerforming EFS download sample. Downloading file into (" + efsFileURIPath + ")");
    
    // Generate the file transfer request that we will be uploading to the system   
    String connectionKey = generateConnectionKey(SERVER_IP, SERVER_PORT_STR, SERVER_USERNAME, SERVER_PASSWORD);
    final TransferFilesRequestModel efsDownloadTransferRequest = getTransferFilesModel(connectionKey, "local");
    efsDownloadTransferRequest.setLocalWorkingDirectory(efsFileURIPath.toString());
    
    List<String> filesToDownload = Arrays.asList(SERVER_FILE_TO_DOWNLOAD);
    efsDownloadTransferRequest.setFileList(filesToDownload);
    
    // Submit the download request to the TransferAgent
    performTAFileTransfer(efsDownloadTransferRequest);
    System.out.println("Download operation has completed.");
  }
  
  public static void main (String[] args) throws Exception {
    TAEFSSample sampleApp = new TAEFSSample();
    sampleApp.performEFSListingExample();
    sampleApp.performEFSUploadExample();
    sampleApp.performEFSDownloadExample();
  }
}
