package samples.transferAgent;
import javax.ws.rs.core.Response.Status;

import unlimited.fc.rest.shared.model.FileObjectsModel;

/**
 * Small sample that performs standard file listing operations against a TransferAgent that resides
 * on an instance separate from the resource executing the sample.
 * 
 * Sample prerequisite:
 *    --- TransferAgent installed with the following:
 *      --- License with the "Remote Control" feature enabled
 *      --- Remote control username and password set to the values of your choosing
 *      
 *    --- Update TransferAgent remote connection settings with the connection information for your system
 *    --- Update the Server connection settings in TransferAgentSampleConstants.java to point to the Server you wish to connect to
 */
public class TARemoteControlSample extends TASampleBase {

  // Remote control TransferAgent connection settings
  private final static String REMOTE_TA_HOST = "127.0.0.1";
  private final static int REMOTE_TA_PORT = 12680;
  private final static String REMOTE_CONTROL_USERNAME = "admin";
  private final static String REMOTE_CONTROL_PASSWORD = "system";
  
  // Default constructor for the remote control sample
  public TARemoteControlSample() {
    super();
    setupRESTEvokerForRemoteControl();
  }
  
  /**
   * Helper method that establishes the credentials needed to access the remote TransferAgent instance.
   * 
   * Once the username and password values have been set, the RESTEvoker will append a "RESTAuthorization"
   * header to all requests made by this sample. By setting this header, the sample will be able to successfully
   * supply a set of credentials, that the TransferAgent will validate upon receiving. Once the credentials have been
   * successfully validated, the call be allowed through, and it will respond with the corresponding result data.  
   */
  private void setupRESTEvokerForRemoteControl() {
    System.out.println("Applying settings required for the remote control access");
    restEvoker.setHost(REMOTE_TA_HOST);
    restEvoker.setPort(REMOTE_TA_PORT);
    restEvoker.setUsername(REMOTE_CONTROL_USERNAME);
    restEvoker.setPassword(REMOTE_CONTROL_PASSWORD);
  }
  
  /**
   * Retrieves the roots of the local file system that the remote TransferAgent resides on
   * @throws Exception Thrown if the REST service returns with unexpected status code.
   */
  public void performLocalFileListSample() throws Exception {
    System.out.println("\nPerforming local file listing of remote TransferAgent instance");
    String localRESTPath = getRESTURLForPath(LOCAL_PATH);
    FileObjectsModel localFileListing = restEvoker.doGETCall(localRESTPath, FileObjectsModel.class, Status.OK);
    outputFileListData(localFileListing);
  }
  
  /**
   * Retrieves the file listing of the Server that the remote TransferAgent is pointing to.
   * @throws Exception Thrown if a REST service returns within unexpected status code
   */
  public void performRemoteFileListSample() throws Exception {
    System.out.println("\nPerforming remote file listing of files located at (" + SERVER_IP + ")");
    
    String remoteRESTPath = getRESTURLForPath(REMOTE_PATH);
    String serverConnectionKey = generateConnectionKey(SERVER_IP, SERVER_PORT_STR, SERVER_USERNAME, SERVER_PASSWORD);    
    restEvoker.addQueryParam("connectionKey", serverConnectionKey);
    
    FileObjectsModel remoteFileData = restEvoker.doGETCall(remoteRESTPath, FileObjectsModel.class, Status.OK);
    outputFileListData(remoteFileData);
    System.out.println("Remote file listing has completed");
  }
  
  public static void main (String[] args) throws Exception {
    TARemoteControlSample sampleApp = new TARemoteControlSample();    
    sampleApp.performLocalFileListSample();
    sampleApp.performRemoteFileListSample();
  }
}
