package samples.transferAgentViaCentral;
import java.util.List;

import javax.ws.rs.core.Response.Status;

import samples.SampleBase;
import unlimited.fc.rest.central.model.CentralLinkModel;
import unlimited.fc.rest.central.model.NodeModel;
import unlimited.fc.rest.central.model.NodesModel;
import unlimited.fc.rest.central.type.ConnectionStatusType;
import util.RESTEvoker;

/**
 * This base sample class that illustrates using the TransferAgent application through FileCatalyst
 * Central's REST proxy dispatching system. The sample performs this by first establishing a "baseURL" that
 * is used to build the beginning of each of the TransferAgent calls made through the REST Proxy. 
 * 
 * Once the "baseURL" is generated, the sample can perform remote calls to the TransferAgent by simply 
 * adjusting the syntax of the command that would have been executed. Instead of calling /rs/files/local/
 * to get the file listing, the sample will call baseURL + "/files/local". This will cause the call to go
 * through Central, to the TransferAgent, where it will then be performed.
 * 
 * Note: All REST services available through the TransferAgent's REST API should be available through the
 * Central REST proxy. For more information regarding the available TransferAgent REST services, please visit
 * the REST API documentation for the TransferAgent application.  
 */
public class TAViaCentralSampleBase extends SampleBase implements TAViaCentralConstants {
  
  // Instance variables
  private String taViaCentralRESTBase;

  protected TAViaCentralSampleBase() throws Exception {
    super();
    this.restEvoker = new RESTEvoker(CENTRAL_SCHEME, CENTRAL_IP, CENTRAL_PORT);
    this.restEvoker.setUsername(CENTRAL_USERNAME);
    this.restEvoker.setPassword(CENTRAL_PASSWORD);
    this.restEvoker.setDebug(false);
    
    taViaCentralRESTBase = generateTAviaCentralRESTBase();
  }
  
  
  /**
   * This method generates the base REST URL information that will be used for all TransferAgent REST calls that go through
   * Central. The method does by retrieving the list of all available nodes from Central, finding the first active TransferAgent 
   * node, and then constructing a string containing the necessary information for communicating with the TransferAgent via Central
   *  
   * @return A string representing the base URL that will be used to perform REST calls through Central.
   * @throws Exception Thrown if the sample application fails to find a valid TA node, or if it there are problems talking with Central
   */
  private String generateTAviaCentralRESTBase() throws Exception {
    final String taNodeType = "FileCatalyst TransferService";
    final NodesModel centralNodeData = restEvoker.doGETCall(CENTRAL_NODES_REST_PATH, NodesModel.class, Status.OK);

    NodeModel transferAgentNode = null;
    if (centralNodeData.getAaData() != null) {
      for (NodeModel node : centralNodeData.getAaData()) {
        final boolean nodeActive = node.getConnectionStatus() == ConnectionStatusType.ACTIVE;
        final boolean isTransferAgentNode = node.getNodeType().equals(taNodeType);
        if (nodeActive && isTransferAgentNode) {
          transferAgentNode = node;
          break;
        }
      }
    }
    
    // If the system found a valid node within Central's list, then construct the base URL. Otherwise, fail
    if (transferAgentNode != null) {
      outputRESTResourcesForNode(transferAgentNode);
      
      String returnVal = CENTRAL_NODES_REST_PATH + "/" + transferAgentNode.getNodeID() + "/ta/";
      return returnVal;
    } else {
      throw new Exception("Unable to run sample. No TransferAgent node present on the Central");
    }
  }
  
  private void outputRESTResourcesForNode(NodeModel taNode) {
    final List<CentralLinkModel> restResourceData = taNode.getLink();
    if (restResourceData != null) {
      System.out.println("\nHere is the list of known REST services available for the provided node");
      for (CentralLinkModel link: restResourceData) {
        String restPath = link.getRest().toString();
        System.out.format("%-30s%-30s%n", "  ---> REST Path: " + restPath, "Resource type: " + link.getType().name());
      }
      System.out.println();
    }
  }
  
  @Override
  public String getRESTURLForPath(String restPath) {
    return taViaCentralRESTBase + restPath;
  }
}
