package util;
import java.io.IOException;
import java.net.URI;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Base64;
import java.util.Base64.Encoder;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.Invocation.Builder;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.UriBuilder;

import org.glassfish.jersey.client.ClientProperties;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * A small class that encapsulates all of the functionality to evoke REST services, across all
 * available HTTP methods that can be used. 
 * 
 * The underlying system uses the Jersey and Jackson libraries to provide an easy to use shell, which
 * should provide all of the necessary functionality needed for performing REST calls against an external Server 
 */
public class RESTEvoker {

  private int port;
  private String scheme;
  private String host;
  private String username;
  private String password;
  private boolean debug;
  
  private MediaType dataType;
  private Map<String, String> queryParams;
  
  private final static String SSL_CONTEXT_TYPE = "TLS";
  private final static String REST_AUTHORIZATION_HEADER = "RESTAuthorization";
  private final static ObjectMapper dataMapper = createDefaultMapper();

  public RESTEvoker(String scheme, String host, int port) {
    this.host = host;
    this.port = port;
    this.scheme = scheme;
    
    this.queryParams = new HashMap<>();
    this.dataType = MediaType.APPLICATION_JSON_TYPE;
  }

  public String getBaseURL() {
    StringBuilder sb = new StringBuilder(scheme);
    sb.append("://").append(host).append(":").append(port);

    return sb.toString();
  }

  /**
   * Invokes a GET request to the currently configured WebServer.
   * 
   * @param <T> Type of the object that is to be returned by the GET request
   * @param restPath REST entry point of the service you wish to call
   * @param responseClass Instance of the response class that you are expecting to come back
   * @return Instance objecting containing the data from the response
   * @throws Exception Thrown if there is an error communicating with the TA, or if the response status code is
   *           different then expected
   */
  public <T> T doGETCall(String restPath, Class<T> responseClass, Status expectedStatus) throws Exception {
    final Builder builder = initWebTargetBuilder(restPath);
    printDebug("Evoking GET request to the (" + restPath + ") end point");
    
    final Response restResponse = builder.get();
    final String responseDoc = restResponse.readEntity(String.class);
    printDebug("GET request to (" + restPath + ") returned. Status = " + restResponse.getStatus() + ", response = " + responseDoc);

    validateRESTResponse(restResponse, responseDoc, expectedStatus);
    final T returnVal = mapResponseToData(responseDoc, responseClass);
    return returnVal;
  }

  
  /**
   * Invokes a POST request to the currently configured WebServer.
   * 
   * @param <T> Type of the object that is to be returned by the POST request
   * @param restPath REST entry point of the service you wish to call
   * @param responseClass Instance of the response class that you are expecting to come back
   * @return Instance objecting containing the data from the response
   * @throws Exception Thrown if there is an error communicating with the TA, or if the response status code is
   *           different then expected
   */
  public <T> T doPOSTCall(String restPath, Class<T> responseClass, Status expectedStatus) throws Exception {
    return doPOSTCall(restPath, null, responseClass, expectedStatus);
  }
  
  /**
   * Invokes a POST request to the currently configured WebServer. 
   * @param <T> Type of the payload that you are going to be providing
   * @param <R> Return type of the POST call that will be executed
   * @param restPath REST entry point of the service you wish to call
   * @param restPayload Data that will be submitted within the payload of the REST call
   * @param responseClass Class of the response that you are expecting to come back
   * @return Class instance containing the data from the response
   * @throws Exception Thrown if there is an error communicating with the TA, or if the response status code is different then expected 
   */
  public <T, R> R doPOSTCall(String restPath, T restPayload, Class<R> responseClass, Status expectedStatus) throws Exception {
    final Builder builder = initWebTargetBuilder(restPath);
    final Entity<?> payload = (restPayload != null) ? Entity.entity(restPayload, dataType) 
                                                    : Entity.entity("", dataType);
    
    printDebug("Evoking POST request to the (" + restPath + ") end point");
    final Response restResponse = builder.post(payload);
    final String responseDoc = restResponse.readEntity(String.class);
    printDebug("POST request to (" + restPath + ") returned. Status = " + restResponse.getStatus() + ", response = " + responseDoc);

    validateRESTResponse(restResponse, responseDoc, expectedStatus);
    final R returnVal = mapResponseToData(responseDoc, responseClass);
    return returnVal;
  }

  /**
   * Invokes a DELETE request to the currently configured WebServer
   * @param <T> Type of object that is to be returned by the DELETE request 
   * @param restPath REST entry point of the Service to be called
   * @param responseClass Response class that is expected
   * @return Instance of expected return class
   * @throws Exception Thrown if there is an error communicating with the TA, or if the response status code is different then expected
   */
  public <T> T doDELETECall(String restPath, Class<T> responseClass, Status expectedStatus) throws Exception {
    final Builder builder = initWebTargetBuilder(restPath);
    printDebug("Evoking DELETE request to the (" + restPath + ") end point");
    
    final Response restResponse = builder.delete();
    final String responseDoc = restResponse.readEntity(String.class);
    printDebug("DELETE request to (" + restPath + ") returned. Status = " + restResponse.getStatus() + ", response = " + responseDoc);

    validateRESTResponse(restResponse, responseDoc, expectedStatus);
    final T returnVal = mapResponseToData(responseDoc, responseClass);
    return returnVal;
  }
  
  public void addQueryParam(String queryKey, String queryVal) {
    queryParams.put(queryKey, queryVal);
  }
  
  public void removeQueryParam(String queryKey) {
    queryParams.remove(queryKey);
  }
  
  /**
   * @return the port
   */
  public int getPort() {
    return port;
  }

  /**
   * @param port the port to set
   */
  public void setPort(int port) {
    this.port = port;
  }

  /**
   * @return the scheme
   */
  public String getScheme() {
    return scheme;
  }

  /**
   * @param scheme the scheme to set
   */
  public void setScheme(String scheme) {
    this.scheme = scheme;
  }

  /**
   * @return the host
   */
  public String getHost() {
    return host;
  }

  /**
   * @param host the host to set
   */
  public void setHost(String host) {
    this.host = host;
  }

  /**
   * @return the username
   */
  public String getUsername() {
    return username;
  }

  /**
   * @param username the username to set
   */
  public void setUsername(String username) {
    this.username = username;
  }

  /**
   * @return the password
   */
  public String getPassword() {
    return password;
  }

  /**
   * @param password the password to set
   */
  public void setPassword(String password) {
    this.password = password;
  }

  /**
   * @return the dataType
   */
  public MediaType getDataType() {
    return dataType;
  }

  /**
   * @param dataType the dataType to set
   */
  public void setDataType(MediaType dataType) {
    this.dataType = dataType;
  }

  /**
   * @return the debug
   */
  public boolean isDebug() {
    return debug;
  }

  /**
   * @param debug the debug to set
   */
  public void setDebug(boolean debug) {
    this.debug = debug;
  }

  private void printDebug(String message) {
    if (debug) {
      System.out.println("## DEBUG: " + message);
    }
  }
  
  private void validateRESTResponse(Response response, String responseDoc, Status expectedStatus) throws Exception {
    if (response.getStatus() != expectedStatus.getStatusCode()) {
      final Status responseStatus = Status.fromStatusCode(response.getStatus());
      final String errorMsg = "Invalid status (" + responseStatus + ") received. Response: " + responseDoc;
      throw new Exception(errorMsg);
    }
  }
  
  private Builder initWebTargetBuilder(String restPath) {
    final WebTarget restTarget = getWebResource(restPath);
    
    final Builder builder = restTarget.request(dataType);
    builder.header(REST_AUTHORIZATION_HEADER, getRESTAuthorization());
    return builder;
  }
  
  private WebTarget getWebResource(String path) {
    Client client = ClientBuilder.newBuilder()
                                 .sslContext(generateSSLContext())
                                 .property(ClientProperties.SUPPRESS_HTTP_COMPLIANCE_VALIDATION, true)
                                 .hostnameVerifier((host, session) -> true)
                                 .build();
    
    String editedPath = path.replaceAll(" ", "%20");
    
    UriBuilder url = UriBuilder.fromUri(getBaseURL() + editedPath);
    queryParams.forEach((key, value) -> url.queryParam(key, value));
    queryParams.clear();
    
    final URI build = url.build();
    printDebug("Constructed REST connection for: " + url);
    
    WebTarget WebTarget = client.target(build);
    return WebTarget;
  }

  private SSLContext generateSSLContext() {
    TrustManager[] certs = getTrustManagerCerrts();
    SSLContext ctx = null;
    try {
      ctx = SSLContext.getInstance(SSL_CONTEXT_TYPE);
      ctx.init(null, certs, new SecureRandom());
    } catch (java.security.GeneralSecurityException ex) {}
    HttpsURLConnection.setDefaultSSLSocketFactory(ctx.getSocketFactory());
    return ctx;
  }

  private TrustManager[] getTrustManagerCerrts() {
    TrustManager[] certs = new TrustManager[] { new X509TrustManager() {

      @Override
      public X509Certificate[] getAcceptedIssuers() {
        return null;
      }

      @Override
      public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException {}

      @Override
      public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException {}
    } };
    return certs;
  }
  
  private String getRESTAuthorization() {
    Encoder dataEncoder = Base64.getEncoder();
    
    StringBuilder sb = new StringBuilder();
    sb.append(Optional.ofNullable(username).orElse(""))
      .append(":")
      .append(Optional.ofNullable(password).orElse(""));
    
    String rawData = sb.toString();
    String encodedData = new String(dataEncoder.encode(rawData.getBytes()));
    return encodedData;
  }
  
  private static ObjectMapper createDefaultMapper() {
    ObjectMapper result = new ObjectMapper();
    result.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false).configure(MapperFeature.SORT_PROPERTIES_ALPHABETICALLY, true);
    return result;
  }
  
  private <T> T mapResponseToData(String responseData, Class<T> returnClazz) throws IOException {
    final T returnVal = dataMapper.readValue(responseData, returnClazz);
    return returnVal;
  }
}
